#include "endgame.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct {
  move_t move;
  int previous;
  int index;
} bfs_item_t;

typedef struct {
  bfs_item_t *l;
  int size;
  int maxsize;
  int next;
} bfs_t;

static void bfs_init(bfs_t *b)
{
  b->l = NULL;
  b->maxsize = b->size = 0;
  b->next = 0;
}

static void bfs_free(bfs_t *b)
{
  free(b->l);
}

static int bfs_over(bfs_t *b)
{
  return b->next >= b->size;
}

static bfs_item_t bfs_get(bfs_t *b)
{
  bfs_item_t ret = b->l[b->next];
  b->next++;
  return ret;
}

static void bfs_add(bfs_t *b, move_t m, int previous)
{
  if (b->size == b->maxsize) {
    b->maxsize += 4096;
    b->l = realloc(b->l, b->maxsize * sizeof(bfs_item_t)); if (b->l == NULL) abort();
  }
  int pos = b->size;
  b->l[pos] = (bfs_item_t){.move = m, .previous = previous, .index = pos};
  b->size++;
}

static void bfs_dump(bfs_t *b)
{
  printf("bfs[%d].l = ", b->size);
  for (int i = 0; i < b->size; i++) printf(" [%d %d]", b->l[i].index, b->l[i].previous);
  printf("\n");
}

#define set_grid_and_position(grid, pos, v) do { \
    grid[(int)pos.y][(int)pos.x] = v; \
  } while (0)
#define get_working_grid(p, pos) grid[(int)pos.y][(int)pos.x]
#define DO_MOVE(p, m) do { \
    point_t pt[3]; \
    positions_around(pt, (point_t){.x = (m).x, .y = (m).y}); \
    if ((m).direction == 0) { \
      /* left */ \
      int v = get_working_grid(p, pt[0]); \
      set_grid_and_position(p, pt[0], get_working_grid(p, pt[2])); \
      set_grid_and_position(p, pt[2], get_working_grid(p, pt[1])); \
      set_grid_and_position(p, pt[1], v); \
    } else { \
      /* right */ \
      int v = get_working_grid(p, pt[0]); \
      set_grid_and_position(p, pt[0], get_working_grid(p, pt[1])); \
      set_grid_and_position(p, pt[1], get_working_grid(p, pt[2])); \
      set_grid_and_position(p, pt[2], v); \
    } \
  } while (0)
#define UNDO_MOVE(p, m) do { \
    point_t pt[3]; \
    positions_around(pt, (point_t){.x = (m).x, .y = (m).y}); \
    if ((m).direction == 1) { \
      /* undo right is do left */ \
      int v = get_working_grid(p, pt[0]); \
      set_grid_and_position(p, pt[0], get_working_grid(p, pt[2])); \
      set_grid_and_position(p, pt[2], get_working_grid(p, pt[1])); \
      set_grid_and_position(p, pt[1], v); \
    } else { \
      /* undo left is do right */ \
      int v = get_working_grid(p, pt[0]); \
      set_grid_and_position(p, pt[0], get_working_grid(p, pt[1])); \
      set_grid_and_position(p, pt[1], get_working_grid(p, pt[2])); \
      set_grid_and_position(p, pt[2], v); \
    } \
  } while (0)

typedef struct {
  move_t *moves;
  int size;
} path_t;

/* beware: returned path is in reverse */
static path_t get_path(bfs_t *b, bfs_item_t p)
{
  path_t ret = { .moves = NULL, .size = 0 };
  int index = p.index;
  while (index != -1) {
    if ((ret.size & 4095) == 0) {
      ret.moves = realloc(ret.moves, (ret.size + 4096) * sizeof(move_t)); if (ret.moves == NULL) abort();
    }
    ret.moves[ret.size] = b->l[index].move;
    index = b->l[index].previous;
    ret.size++;
  }
  return ret;
}

static void free_path(path_t p)
{
  free(p.moves);
}

static int do_path(unsigned int grid[2*N-1][2*N-1], bfs_t *b, bfs_item_t p)
{
  path_t path = get_path(b, p);
printf("do_path len %d\n", path.size);

  for (int i = path.size - 1; i >= 0; i--)
    DO_MOVE(grid, path.moves[i]);

  int ret = path.moves[0].direction;

  free_path(path);

  return ret;
}

static void undo_path(unsigned int grid[2*N-1][2*N-1], bfs_t *b, bfs_item_t p)
{
  path_t path = get_path(b, p);

  for (int i = 0; i < path.size; i++)
    UNDO_MOVE(grid, path.moves[i]);

  free_path(path);
}

static moves_t path_to_moves(bfs_t *b, bfs_item_t p)
{
  path_t path = get_path(b, p);
  moves_t ret;

  ret.size = path.size;
  ret.moves = malloc(path.size * sizeof(move_t)); if (ret.moves == NULL) abort();

  for (int i = path.size - 1; i >= 0; i--)
    ret.moves[path.size - 1 - i] = path.moves[i];

  free_path(path);

  return ret;
}

static int get_valid_moves(move_t *m, unsigned int grid[2*N-1][2*N-1], int direction)
{
  int ret = 0;
  point_t moves[8] = {
    { .x = N-1, .y = 2*N-2 },
    { .x = N-1, .y = 2*N-1 },
    { .x = N,   .y = 2*N-1 },
    { .x = N-1, .y = 2*N   },
    { .x = N,   .y = 2*N   },
    { .x = N-1, .y = 2*N+1 },
    { .x = N-1, .y = 2*N+3 },
    { .x = N-1, .y = 2*N+4 },
  };
  for (int i = 0; i < 8; i++) {
    int valid_cursor(point_t p, unsigned int grid[2*N-1][2*N-1]);
    if (valid_cursor(moves[i], grid)) {
      *m = (move_t){ .x = moves[i].x, .y = moves[i].y, .direction = direction };
      m++;
      ret++;
    }
  }
  return ret;
}

static void set_busy(unsigned char *busy, int *state, int direction)
{
  long offset = 0;
  for (int i = 0; i < 10; i++) {
    offset *= 10;
    offset += state[i];
  }
  offset *= 2;
  offset |= direction;
  busy[offset / 8] |= 1 << (offset % 8);
}

static int is_busy(unsigned char *busy, int *state, int direction)
{
  long offset = 0;
  for (int i = 0; i < 10; i++) {
    offset *= 10;
    offset += state[i];
  }
  offset *= 2;
  offset |= direction;
  return busy[offset / 8] & (1 << (offset % 8));
}

static void get_state(int stone_to_state[(2*N-1)*(2*N-1)],
                      unsigned int grid[2*N-1][2*N-1],
                      position_t grid_pos[10],
                      int ret_state[10])
{
  for (int i = 0; i < 10; i++) {
    int x = grid_pos[i].x;
    int y = grid_pos[i].y;
    int stone = grid[y][x];
    int state = stone_to_state[stone];
    ret_state[i] = state;
  }
}

void dump(unsigned int working_grid[2*N-1][2*N-1]);

static int last_ring_solvable(unsigned int grid[2*N-1][2*N-1], int start_direction)
{
  moves_t get_last_moves(unsigned int grid[2*N-1][2*N-1], int start_direction);
  moves_t moves = get_last_moves(grid, start_direction);
  free(moves.moves);
  return moves.size >= 0;
}

moves_t endgame(unsigned int grid[2*N-1][2*N-1], int start_direction)
{
dump(grid);
  moves_t ret = { .moves = NULL, .size = 0 };
  bfs_t b;
  bfs_init(&b);
  unsigned char *busy = malloc((long)2*10*10*10*10*10*10*10*10*10*10/8); if (busy == NULL) return (moves_t){.size = 0, .moves = NULL};

  /* init stone_to_state */
  int stone_to_state[(2*N-1)*(2*N-1)];
  position_t grid_pos[10] = {
    {.x = N-2, .y = N-2 },
    {.x = N-1, .y = N-2 },
    {.x = N-2, .y = N-1 },
    {.x = N-1, .y = N-1 },
    {.x = N,   .y = N-1 },
    {.x = N-2, .y = N },
    {.x = N-1, .y = N },
    {.x = N-2, .y = N+1 },
    {.x = N-1, .y = N+1 },
    {.x = N-2, .y = N+2 },
  };
  /* canonize grid_pos */
  for (int i = 0; i < 10; i++) {
    point_t p = canonical_position((point_t){.x = grid_pos[i].x, .y = grid_pos[i].y });
    grid_pos[i].x = p.x;
    grid_pos[i].y = p.y;
    printf("grid_pos[%d] = %d %d\n", i, grid_pos[i].x, grid_pos[i].y);
  }
  /* get stone_to_state */
  for (int i = 0; i < 10; i++)
    stone_to_state[grid[grid_pos[i].y][grid_pos[i].x]] = i;
  /* get initial 3 stones not in the ring */
  int unring_stones[3];
  for (int i = 0; i < 3; i++) {
    int x = grid_pos[i + 7].x;
    int y = grid_pos[i + 7].y;
    unring_stones[i] = grid[y][x];
  }

  int direction = start_direction;
  move_t moves[8];

  int count = get_valid_moves(moves, grid, direction);
  for (int i = 0; i < count; i++)
    bfs_add(&b, moves[i], -1);

  direction = 1 - direction;

  while (!bfs_over(&b)) {
    bfs_item_t p = bfs_get(&b);
printf("do %d\n", p.index);
    int last_direction = do_path(grid, &b, p);
    if (last_ring_solvable(grid, 1 - last_direction)) {
      /* check last 3 stones - must be at init pos */
      int unring_end[3];
      for (int i = 0; i < 3; i++) {
        int x = grid_pos[i + 7].x;
        int y = grid_pos[i + 7].y;
        unring_end[i] = grid[y][x];
      }
      if (memcpy(unring_stones, unring_end, 3 * sizeof(int))) {
        undo_path(grid, &b, p);
        ret = path_to_moves(&b, p);
void dump_moves(moves_t *);
printf("returned moves[%d]\n", ret.size);
dump_moves(&ret);
printf("\n");
        break;
      }
    }
    count = get_valid_moves(moves, grid, 1 - last_direction);
    for (int i = 0; i < count; i++) {
      DO_MOVE(grid, moves[i]);
      int state[10];
      get_state(stone_to_state, grid, grid_pos, state);
      if (!is_busy(busy, state, 1 - last_direction)) {
        bfs_add(&b, moves[i], p.index);
        set_busy(busy, state, 1 - last_direction);
      }
      UNDO_MOVE(grid, moves[i]);
    }
    undo_path(grid, &b, p);
  }

  bfs_free(&b);
  free(busy);

  return ret;
}
