#include "last-ring.h"

#include <stdio.h>
#include <stdlib.h>

typedef struct {
  int count;  /* -1 if no path exists */
  int *moves; /* moves stored as [1..6] for clock direction
               * and [-1..-6] for counterclock
               */
} moves_list_t;

extern moves_list_t last_ring[2][5040];

int pack_state(int state[7])
{
  return state[0] |
         (state[1] << 3) |
         (state[2] << 6) |
         (state[3] << 9) |
         (state[4] << 12) |
         (state[5] << 15) |
         (state[6] << 18);
}

void unpack_state(int state[7], int in)
{
  state[0] = in & 0x07;
  state[1] = (in >> 3) & 0x07;
  state[2] = (in >> 6) & 0x07;
  state[3] = (in >> 9) & 0x07;
  state[4] = (in >> 12) & 0x07;
  state[5] = (in >> 15) & 0x07;
  state[6] = (in >> 18) & 0x07;
}

int packed_state_to_move_index[8*8*8*8*8*8*8];
int move_index_to_packed_state[5040];

void init_move_index(void)
{
  static int inited = 0;
  if (inited) return;
  inited = 1;

  int move_index = 0;
  for (int packed_state = 0; packed_state < 2097152; packed_state++)
    packed_state_to_move_index[packed_state] = -1;
  for (int packed_state = 0; packed_state < 2097152; packed_state++) {
    int state[7];
    unpack_state(state, packed_state);
    /* all values must be in [0..6] */
    if (state[0] > 6 || state[1] > 6 || state[2] > 6 || state[3] > 6
        || state[4] > 6 || state[5] > 6 || state[6] > 6)
      continue;
    /* all must be != */
    int v = 0;
    for (int i = 0; i < 7; i++) v |= 1 << state[i];
    if (v != 127) continue;
    packed_state_to_move_index[packed_state] = move_index;
    move_index_to_packed_state[move_index] = packed_state;
    move_index++;
  }
  printf("last move_index %d\n", move_index);
}

moves_t get_last_moves(unsigned int grid[2*N-1][2*N-1], int start_direction)
{
  init_move_index();

  /* get solution grid */
  int solution_grid[2 * N - 1][2 * N - 1];
  int stone = 1;
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    /* skip center grid position */
    if (is_grid(y, x) && !(y == N - 1 && x == N - 1)) {
      solution_grid[y][x] = stone;
      stone++;
    }
  solution_grid[N - 1][N - 1] = 0;


  /* last ring solver works with stones names 0..6
   * we need to translate from grid to last ring
   * see generate-last-ring.c for last ring order
   * (3 is center, so is stone 0)
   */
  int stone_id[7];
  stone_id[0] = solution_grid[N - 2][N - 2];
  stone_id[1] = solution_grid[N - 2][N - 1];
  stone_id[2] = solution_grid[N - 1][N - 2];
  stone_id[3] = solution_grid[N - 1][N - 1];
  stone_id[4] = solution_grid[N - 1][N    ];
  stone_id[5] = solution_grid[N    ][N - 2];
  stone_id[6] = solution_grid[N    ][N - 1];

printf("stones %d %d %d %d %d %d %d\n", stone_id[0], stone_id[1], stone_id[2], stone_id[3], stone_id[4], stone_id[5], stone_id[6]);

  /* get initial position */
  int initial_position[7] = { -1, -1, -1, -1, -1, -1, -1};
  int pos_to_grid_x[7] = { N-2, N-1, N-2, N-1, N, N-2, N-1 };
  int pos_to_grid_y[7] = { N-2, N-2, N-1, N-1, N-1, N, N };
  for (int pos = 0; pos < 7; pos++) {
    int stone = grid[pos_to_grid_y[pos]][pos_to_grid_x[pos]];
    for (int tpos = 0; tpos < 7; tpos++)
      if (stone == stone_id[tpos])
        initial_position[pos] = tpos;
    printf("%d ", stone);
  }
  printf("\n");
printf("init pos %d %d %d %d %d %d %d\n", initial_position[0], initial_position[1], initial_position[2], initial_position[3], initial_position[4], initial_position[5], initial_position[6]);
  for (int pos = 0; pos < 7; pos++)
    if (initial_position[pos] == -1) {
      printf("bad init pos\n");
      return (moves_t) { .size = -1, .moves = (void *)0 };
    }

  int packed_state = pack_state(initial_position);

  moves_list_t *move_list = &last_ring[start_direction][packed_state_to_move_index[packed_state]];

  if (move_list->count == -1) { printf("t: not solvable\n"); return (moves_t) { .size = -1, .moves = (void *)0 }; }

  printf("last ring moves[%d]:", move_list->count);
  for (int i = 0; i < move_list->count; i++) printf(" %d", move_list->moves[i]);
  printf("\n");

  moves_t ret;
  ret.size = move_list->count;
  ret.moves = malloc(ret.size * sizeof(move_t));
  if (ret.moves == NULL) {
    printf("error: could not allocate memory, no moves will be returned\n");
    return (moves_t) { .size = 0, .moves = (void *)0 };
  }

  for (int i = 0; i < move_list->count; i++) {
    ret.moves[i].direction = move_list->moves[i] > 0 ? 0 : 1;
    switch (abs(move_list->moves[i])) {
    case 1: ret.moves[i].x = N-1; ret.moves[i].y = 2*N-2; break;
    case 2: ret.moves[i].x = N  ; ret.moves[i].y = 2*N-1; break;
    case 3: ret.moves[i].x = N  ; ret.moves[i].y = 2*N  ; break;
    case 4: ret.moves[i].x = N-1; ret.moves[i].y = 2*N+1; break;
    case 5: ret.moves[i].x = N-1; ret.moves[i].y = 2*N  ; break;
    case 6: ret.moves[i].x = N-1; ret.moves[i].y = 2*N-1; break;
    }
  }

  return ret;
}
